@@ -43,6 +43,17 @@ public:
     }
 
     vector<float> operator()(RCover* ar) {
+        PyInit_error_function();
+        vector<float> result;
+        if (pyFunction != nullptr) { // nullptr check
+            float* result_pointer =  call_python_support_error_class_function(pyFunction, ar); // note, no way of checking for errors until you return to Python
+            result.push_back(result_pointer[0]);
+            result.push_back(result_pointer[1]);
+        }
+        return result;
+    }
+
+    /*vector<float> operator()(RCover* ar) {
         int status = PyImport_AppendInittab("error_function", PyInit_error_function);
         if (status == -1) {
             vector<float> result;
@@ -56,7 +67,6 @@ public:
             return result;
         }
 
-//        PyInit_error_function();
         vector<float> result;
         if (pyFunction) { // nullptr check
             result = *call_python_support_error_class_function(pyFunction, ar); // note, no way of checking for errors until you return to Python
@@ -64,7 +74,7 @@ public:
 
         Py_Finalize();
         return result;
-    }
+    }*/
 
 private:
     PyObject* pyFunction;
